﻿namespace AY_emulator
{
    using NAudio.Wave;

    //  The AY-3-8910/8912 contains 16 internal registers as follows:

    //  Register    Function                        Range
    //  0           Channel A fine pitch            8-bit(0-255)
    //  1           Channel A course pitch          4-bit(0-15)
    //  2           Channel B fine pitch            8-bit(0-255)
    //  3           Channel B course pitch          4-bit(0-15)
    //  4           Channel C fine pitch            8-bit(0-255)
    //  5           Channel C course pitch          4-bit(0-15)
    //  6           Noise pitch                     5-bit(0-31)
    //  7           Mixer                           8-bit(see below)
    //  8           Channel A volume                4-bit(0-15, see below)
    //  9           Channel B volume                4-bit(0-15, see below)
    // 10           Channel C volume                4-bit(0-15, see below)
    // 11           Envelope fine duration          8-bit(0-255)
    // 12           Envelope course duration        8-bit(0-255)
    // 13           Envelope shape                  4-bit(0-15)
    // 14           I/O port A                      8-bit(0-255)
    // 15           I/O port B                      8-bit(0-255)

    public class AYChannel
    {
        private const int EnvelopeTimePeriod = 5;
        private const float MaxGain = 0.1f;

        private AYSignalGenerator _noiseGen;
        private AYSignalGenerator _signalGen;
        private MixingWaveProvider32 _mixer;

        private float _playerPitch;
        private float _vol;
        private bool _playerMix, _noisePlayerMix;
        private float _noisePitch;
        private bool _muted;

        private int _envelopePeriod;
        private float _envelopeDuration;
        private byte _envelopeShape;

        public AYChannel(MixingWaveProvider32 mixer)
        {
            _mixer = mixer;
            _playerPitch = 0;
            _vol = 0;
            _playerMix = false;
            _noisePlayerMix = false;
            _envelopeShape = 0;
            _muted = false;

            _signalGen = new AYSignalGenerator(AYSignalGenerator.AYSignalGeneratorType.Square, MaxGain);
            _signalGen.Frequency = _playerPitch;
            _signalGen.Gain = _vol;

            _noiseGen = new AYSignalGenerator(AYSignalGenerator.AYSignalGeneratorType.White, MaxGain);
            _noiseGen.Frequency = _noisePitch;
            _noiseGen.Gain = _vol;

            _mixer.AddInputStream(_signalGen);
            _mixer.AddInputStream(_noiseGen);
        }

        public int TonePeriod
        {
            // Pitch = AY frequency / (16 x Period value).
            // AY frequency = 1.7735 MHz
            get
            {
                return (int)(110844 / _playerPitch);
            }
            set
            {
                if (value != 0)
                {
                    _playerPitch = 110844 / value;
                }
                else
                {
                    _playerPitch = 1;
                }

                _signalGen.Frequency = _playerPitch;
            }
        }

        public int NoisePeriod
        {
            // Noise pitch = AY frequency / (16 x Noise period value).
            // AY frequency = 1.7735 MHz

            get
            {
                return (int)(110844 / _noisePitch);
            }
            set
            {
                if (value != 0)
                {
                    _noisePitch = 110844 / value;
                }

                else
                {
                    _noisePitch = 1;
                }

                _noiseGen.Frequency = _noisePitch;
            }
        }

        public int EnvelopePeriod

        // From the AY-3-8910 Specification:
        // FE = FC/256xEP
        // Where:
        // FE = Envelope frequency
        // FC = Clock frequency
        // EP = Envelope period value
        // Since in general F=1/P, P (in seconds) = 1/F = 256xEP/FC, where P is envelope duration.
        // FC = 1.7735 MHz
        // P = 0.00014435 x EP

        {
            get
            {
                return _envelopePeriod;
            }
            set
            {
                _envelopePeriod = value;
                _envelopeDuration = 0.00014435f * _envelopePeriod;

                _signalGen.EnvelopePeriod = _envelopeDuration;
                _noiseGen.EnvelopePeriod = _envelopeDuration;
            }
        }

        public int Volume
        // Volume can be set to 0-15 but internally it has to be translated to 0-1f.
        // TODO: Handle value == 0 --> Turn off channel.
        {
            get
            {
                return (int)(_vol / (MaxGain * 15));
            }
            set
            {

                // Check if envelope should be switched on or off.
                if ((value & 16) == 16)
                {
                    _signalGen.EnvelopeStart();
                    _noiseGen.EnvelopeStart();
                }
                else
                {
                    _signalGen.EnvelopeStop();
                    _noiseGen.EnvelopeStop();
                    // Volume = 1-15. _vol = 0-MaxGain.
                    _vol = (MaxGain * value) / 15;
                    _signalGen.Gain = _vol;
                    _noiseGen.Gain = _vol;
                }
            }
        }

        public bool PlayerMix
        {
            get
            {
                return _playerMix;
            }

            set
            {
                {
                    _playerMix = value;
                    _signalGen.Mix = value;
                }
            }
        }

        public bool NoisePlayerMix
        {
            get
            {
                return _noisePlayerMix;
            }

            set
            {
                _noisePlayerMix = value;
                _noiseGen.Mix = value;
            }
        }

        public byte EnvelopeShape
        {
            get
            {
                return _envelopeShape;
            }

            set
            {
                _envelopeShape = value;
                _signalGen.EnvelopeShape = value;
                _noiseGen.EnvelopeShape = value;
            }
        }

        public void MuteChannel()
        {
            if (_muted == false)
            {
                _muted = true;
                _signalGen.Muted = true;
                _noiseGen.Muted = true;
            }
        }

        public void UnmuteChannel()
        {
            if (_muted == true)
            {
                _muted = false;
                _signalGen.Muted = false;
                _noiseGen.Muted = false;
            }
        }
    }
}
