﻿namespace AY_emulator
{
    using System;
    using NAudio.Wave;

    // Credits:
    // Based on/inspired by https://github.com/bencentra/naudio-sinegen/blob/master/naudio-sinegen/SineWaveProvider32.cs by Ben Centra
    // and the SignalGenerator class in NAudio by Mark Heath.

    // This class generates the AY-3-8912 square wave tone and white noise (or pink, if preferred).
    // It is based on the NAudio WaveProvider32 base class, which generate a 32 bit floating point wave stream.
    public class AYSignalGenerator : WaveProvider32
    {
        // Some noise stuff
        private Random _random;
        private float _noiseValue1;
        private float _noiseValue2;

        private AYSignalGeneratorType _type;

        // Private properties for controlling the generated signal.
        private float _gain;
        private float _frequency;
        private bool _mix;

        private int sampleRate;
        float phaseAngleIncrement;


        // Envelope variables
        private float _initialGain;
        private float _maxGain;
        private float _envelopePeriod;
        private byte _envelopeShape;
        private float _envelopeStep;
        private bool _envelopeOn;

        // Current angle in radians for wave function
        private float _phaseAngle;
        private float _noisePhaseAngle;

        // Constructor
        public AYSignalGenerator(AYSignalGeneratorType type, float maxGain)
        {
            _mix = false;
            _phaseAngle = 0;
            _noisePhaseAngle = 0;
            _random = new Random();
            _type = type;
            _frequency = 1;
            _gain = 0;
            _maxGain = maxGain;
            _initialGain = 0;
            _envelopeOn = false;
            Muted = false;
            sampleRate = WaveFormat.SampleRate;
            phaseAngleIncrement = (float)(2 * Math.PI * _frequency / sampleRate);
            if (phaseAngleIncrement > Math.PI)
                phaseAngleIncrement = (float)Math.PI;

        }

        public enum AYSignalGeneratorType
        {
            White,
            Square
        }

        // This is used (indirectly) by the main window to mute/unmute all AY-sound.
        public bool Muted { get; set; }

        // Getters and setters for controlling the generated signal.
        // I haven't found much information on what triggers a reset of the envelope cycle, but i assume that
        // updating Frequency, Envelope shape and Envelope period triggers a reset.
        public float Frequency
        {
            get
            {
                return _frequency;
            }

            set
            {
                _frequency = value;
                phaseAngleIncrement = (float)(2 * Math.PI * _frequency / sampleRate);
                if (phaseAngleIncrement > Math.PI)
                    phaseAngleIncrement = (float)Math.PI;
                EnvelopeReset();
            }
        }

        public byte EnvelopeShape
        {
            get
            {
                return _envelopeShape;
            }

            set
            {
                _envelopeShape = value;
                EnvelopeReset(); //-- Needs to be here for Agent X2.
            }
        }

        public float EnvelopePeriod
        {
            get
            {
                return _envelopePeriod;
            }

            set
            {
                _envelopePeriod = value;
                EnvelopeReset(); //-- See above
            }
        }

        public float Gain
        {
            get
            {
                return _gain;
            }

            set
            {
                _gain = value;
            }
        }

        public bool Mix
        {
            get
            {
                return _mix;
            }

            set
            {
                _mix = value;
                if (_mix)
                {
                    EnvelopeReset();
                    _noiseValue1 = (float)_random.NextDouble();
                    _noiseValue2 = (float)_random.NextDouble();
                }
                
            }
        }
             
        // Start envelope
        public void EnvelopeStart()
        {
            _envelopeOn = true;
            EnvelopeReset();
        }

        // Stop envelope
        public void EnvelopeStop()
        {
            _envelopeOn = false;
        }

        public void EnvelopeReset()
        {
            _envelopeStep = _maxGain / (_envelopePeriod * 44100);

            switch (_envelopeShape)
            {
                case 0: // Shape: \__________
                    _initialGain = _maxGain;
                    _envelopeStep = -_envelopeStep;
                    break;
                case 1: // Shape: \__________
                    _initialGain = _maxGain;
                    _envelopeStep = -_envelopeStep;
                    break;
                case 2: // Shape: \__________
                    _initialGain = _maxGain;
                    _envelopeStep = -_envelopeStep;
                    break;
                case 3: // Shape: \__________
                    _initialGain = _maxGain;
                    _envelopeStep = -_envelopeStep;
                    break;
                case 4: // Shape: /|_________
                    _initialGain = 0;
                    break;
                case 5: // Shape: /|_________
                    _initialGain = 0;
                    break;
                case 6: // Shape: /|_________
                    _initialGain = 0;
                    break;
                case 7: // Shape: /|_________
                    _initialGain = 0;
                    break;
                case 8: // Shape: \|\|\|\|\|\
                    _initialGain = _maxGain;
                    _envelopeStep = -_envelopeStep;
                    break;
                case 9: // Shape: \__________
                    _initialGain = _maxGain;
                    _envelopeStep = -_envelopeStep;
                    break;
                case 10: // Shape: \/\/\/\/\/\
                    _initialGain = _maxGain;
                    _envelopeStep = -_envelopeStep;
                    break;
                case 11: // Shape: \|^^^^^
                    _initialGain = _maxGain;
                    _envelopeStep = -_envelopeStep;
                    break;
                case 12: // Shape: /|/|/|/|/|/
                    _initialGain = 0;
                    break;
                case 13: // Shape: /^^^^^
                    _initialGain = 0;
                    break;
                case 14: // Shape: /\/\/\/\/\/
                    _initialGain = 0;
                    break;
                case 15: // Shape: /|_________
                    _initialGain = 0;
                    break;
            }

            // This condition is needed to ensure that channels that should be silent don't produce
            // a click when _gain is set to _maxGain. Only channels with active envelopes should be handled.
            if (_envelopeOn)
                _gain = _initialGain;
        }

        // Read method
        public override int Read(float[] buffer, int offset, int sampleCount)
        {
            // Loop through every sample
            for (int n = 0; n < sampleCount; n++)
            {
                if (_mix == true & Muted == false)
                {
                    switch (_type)
                    {
                        case AYSignalGeneratorType.White:
                            if (_phaseAngle < Math.PI)
                                buffer[n + offset] = _gain * _noiseValue1;
                            else
                                buffer[n + offset] = _gain * _noiseValue2;
                            break;
                           
                        case AYSignalGeneratorType.Square:
                            if (_phaseAngle < Math.PI)
                                buffer[n + offset] = _gain;
                            else
                                buffer[n + offset] = -_gain;
                            break;
                    }

                    // Envelope adjustment
                    if (_envelopeOn)
                    {
                        _gain = _gain + _envelopeStep;

                        switch (_envelopeShape)
                        {
                            case 0: // Shape: \__________
                                if (_gain < 0)
                                {
                                    _gain = 0;
                                    _envelopeStep = 0;
                                }
                                break;
                            case 1: // Shape: \__________
                                if (_gain < 0)
                                {
                                    _gain = 0;
                                    _envelopeStep = 0;
                                }
                                break;
                            case 2: // Shape: \__________
                                if (_gain < 0)
                                {
                                    _gain = 0;
                                    _envelopeStep = 0;
                                }
                                break;
                            case 3: // Shape: \__________
                                if (_gain < 0)
                                {
                                    _gain = 0;
                                    _envelopeStep = 0;
                                }
                                break;
                            case 4: // Shape: /|_________
                                if (_gain > _maxGain)
                                {
                                    _gain = 0;
                                    _envelopeStep = 0;
                                }
                                break;
                            case 5: // Shape: /|_________
                                if (_gain > _maxGain)
                                {
                                    _gain = 0;
                                    _envelopeStep = 0;
                                }
                                break;
                            case 6: // Shape: /|_________
                                if (_gain > _maxGain)
                                {
                                    _gain = 0;
                                    _envelopeStep = 0;
                                }
                                break;
                            case 7: // Shape: /|_________
                                if (_gain > _maxGain)
                                {
                                    _gain = 0;
                                    _envelopeStep = 0;
                                }
                                break;
                            case 8: // Shape: \|\|\|\|\|\
                                if (_gain < 0)
                                    _gain = _maxGain;
                                break;
                            case 9: // Shape: \__________
                                if (_gain < 0)
                                {
                                    _gain = 0;
                                    _envelopeStep = 0;
                                }
                                break;
                            case 10: // Shape: \/\/\/\/\/\
                                if (_gain < 0)
                                {
                                    _gain = 0;
                                    _envelopeStep = -_envelopeStep;
                                }
                                else if (_gain > _maxGain)
                                {
                                    _gain = _maxGain;
                                    _envelopeStep = -_envelopeStep;
                                }
                                break;
                            case 11: // Shape: \|^^^^^
                                if (_gain < 0)
                                {
                                    _gain = _maxGain;
                                    _envelopeStep = 0;
                                }
                                break;
                            case 12: // Shape: /|/|/|/|/|/
                                if (_gain > _maxGain)
                                    _gain = 0;
                                break;
                            case 13: // Shape: /^^^^^
                                if (_gain > _maxGain)
                                {
                                    _gain = _maxGain;
                                    _envelopeStep = 0;
                                }
                                break;
                            case 14: // Shape: /\/\/\/\/\/
                                if (_gain < 0)
                                {
                                    _gain = 0;
                                    _envelopeStep = -_envelopeStep;
                                }
                                else if (_gain > _maxGain)
                                {
                                    _gain = _maxGain;
                                    _envelopeStep = -_envelopeStep;
                                }
                                break;
                            case 15: // Shape: /|_________
                                if (_gain > _maxGain)
                                {
                                    _gain = 0;
                                    _envelopeStep = 0;
                                }
                                break;
                        }
                    }
                }
                else
                {
                    // Silence
                    buffer[n + offset] = 0;
                }

                _phaseAngle += phaseAngleIncrement;

                if (_phaseAngle > Math.PI * 2)
                {
                    _phaseAngle -= (float)(Math.PI * 2);
                    _noiseValue1 = (float)_random.NextDouble();
                    _noiseValue2 = (float)_random.NextDouble();
                }
                    
            }
            return sampleCount;
        }
    }
}

